import os
import sys
import copy
import math
import numpy as np
import torch
import unittest
import tempfile
import atexit
import textwrap

from typing import Optional
from librosa.feature import melspectrogram

from i6_models.primitives.feature_extraction import *


def test_logmel_librosa_compatibility():

    audio = np.asarray(np.random.random((50000)), dtype=np.float32)
    librosa_mel = melspectrogram(
        y=audio,
        sr=16000,
        n_fft=int(0.05 * 16000),
        hop_length=int(0.0125 * 16000),
        win_length=int(0.05 * 16000),
        fmin=60,
        fmax=7600,
        n_mels=80,
    )
    librosa_log_mel = np.log10(np.maximum(librosa_mel, 1e-10))

    fe_cfg = LogMelFeatureExtractionV1Config(
        sample_rate=16000,
        win_size=0.05,
        hop_size=0.0125,
        f_min=60,
        f_max=7600,
        min_amp=1e-10,
        num_filters=80,
        center=True,
    )
    fe = LogMelFeatureExtractionV1(cfg=fe_cfg)
    audio_tensor = torch.unsqueeze(torch.Tensor(audio), 0)  # [B, T]
    audio_length = torch.tensor([50000])  # [B]
    pytorch_log_mel, frame_length = fe(audio_tensor, audio_length)
    librosa_log_mel = torch.tensor(librosa_log_mel).transpose(0, 1)
    assert torch.allclose(librosa_log_mel, pytorch_log_mel, atol=1e-06)


def test_logmel_length():
    fe_center_cfg = LogMelFeatureExtractionV1Config(
        sample_rate=16000,
        win_size=0.05,
        hop_size=0.0125,
        f_min=60,
        f_max=7600,
        min_amp=1e-10,
        num_filters=80,
        center=True,
    )
    fe_center = LogMelFeatureExtractionV1(cfg=fe_center_cfg)
    fe_no_center_cfg = copy.deepcopy(fe_center_cfg)
    fe_no_center_cfg.center = False
    fe_no_center = LogMelFeatureExtractionV1(cfg=fe_no_center_cfg)
    for i in range(10):
        audio_length = int(np.random.randint(10000, 50000))
        audio = np.asarray(np.random.random(audio_length), dtype=np.float32)
        audio_length = torch.tensor(int(audio_length))
        audio_length = torch.unsqueeze(audio_length, 0)
        audio = torch.unsqueeze(torch.tensor(audio), 0)
        mel_center, length_center = fe_center(audio, audio_length)
        assert torch.all(mel_center.size()[1] == length_center)
        mel_no_center, length_no_center = fe_no_center(audio, audio_length)
        assert torch.all(mel_no_center.size()[1] == length_no_center)


def test_rasr_compatible():
    try:
        from i6_core.lib.rasr_cache import FileArchive
    except ImportError:
        raise unittest.SkipTest("i6_core not available")
    try:
        import soundfile
    except ImportError:
        raise unittest.SkipTest("soundfile not available")
    rasr_feature_extractor_bin_path = (
        "/work/tools22/asr/rasr/rasr_onnx_haswell_0623/arch/linux-x86_64-standard/"
        "feature-extraction.linux-x86_64-standard"
    )
    if not os.path.exists(rasr_feature_extractor_bin_path):
        raise unittest.SkipTest("RASR feature-extraction binary not found")

    torch.manual_seed(42)
    wav_file_path = tempfile.mktemp(suffix=".wav", prefix="tmp-i6models-random-audio")
    atexit.register(os.remove, wav_file_path)
    generate_random_speech_like_audio_wav(wav_file_path)
    rasr_feature_cache_path = generate_rasr_feature_cache_from_wav_and_flow(
        rasr_feature_extractor_bin_path,
        wav_file_path,
        textwrap.dedent(
            f"""\
            <node filter="generic-vector-s16-demultiplex" name="demultiplex" track="$(track)"/>
            <link from="samples" to="demultiplex"/>
            <node filter="generic-convert-vector-s16-to-vector-f32" name="convert"/>
            <link from="demultiplex" to="convert"/>
            <node alpha="1.0" filter="signal-preemphasis" name="preemphasis"/>
            <link from="convert" to="preemphasis"/>
            <node filter="signal-window" length="0.025" name="window" shift="0.01" type="hanning"/>
            <link from="preemphasis" to="window"/>
            <node filter="signal-real-fast-fourier-transform" maximum-input-size="0.025" name="fft"/>
            <link from="window" to="fft"/>
            <node filter="signal-vector-alternating-complex-f32-amplitude" name="amplitude-spectrum"/>
            <link from="fft" to="amplitude-spectrum"/>
            <node filter="signal-filterbank" filter-width="70.12402584464985" name="filterbank"
             warp-differential-unit="false" warping-function="mel"/>
            <link from="amplitude-spectrum" to="filterbank"/>
            <node filter="generic-vector-f32-log-plus" name="nonlinear" value="1.175494e-38"/>
            <link from="filterbank" to="nonlinear"/>
            """
        ),
        flow_output_name="nonlinear",
    )

    rasr_cache = FileArchive(rasr_feature_cache_path, must_exists=True)
    print(rasr_cache.file_list())
    print(rasr_cache.read("corpus/recording/1.attribs", "str"))
    time_, rasr_feat = rasr_cache.read("corpus/recording/1", "feat")
    assert len(time_) == len(rasr_feat)
    print("RASR feature len:", len(rasr_feat), "frame 0 times:", time_[0], "frame 0 shape:", rasr_feat[0].shape)
    rasr_feat = torch.tensor(np.stack(rasr_feat, axis=0), dtype=torch.float32)
    print("RASR feature shape:", rasr_feat.shape)

    cfg = RasrCompatibleLogMelFeatureExtractionV1Config(
        sample_rate=16_000,
        win_size=0.025,
        hop_size=0.01,
        min_amp=1.175494e-38,
        num_filters=80,
    )
    feat_extractor = RasrCompatibleLogMelFeatureExtractionV1(cfg)

    # int16 audio is in [2**15, 2**15-1].
    # This is how BlissToPcmHDFJob does it by default:
    # https://github.com/rwth-i6/i6_core/blob/add09a8b640a2ba5928b815fa65f7504242be038/returnn/hdf.py#L207
    # This is also how our standard RASR flow handles it:
    # https://github.com/rwth-i6/i6_models/pull/44#issuecomment-1938264642
    audio, sample_rate = soundfile.read(open(wav_file_path, "rb"), dtype="int16")
    assert sample_rate == cfg.sample_rate
    audio = torch.tensor(audio.astype(np.float32))  # [-2**15, 2**15-1]
    print("raw audio", _torch_repr(audio))

    i6m_feat, _ = feat_extractor(audio.unsqueeze(0), torch.tensor([len(audio)]))
    i6m_feat = i6m_feat.squeeze(0)

    print("i6_models:", _torch_repr(i6m_feat))
    print("RASR:", _torch_repr(rasr_feat))

    torch.testing.assert_close(i6m_feat, rasr_feat, rtol=1e-5, atol=1e-5)


def test_rasr_compatible_raw_audio_samples():
    try:
        from i6_core.lib.rasr_cache import FileArchive
    except ImportError:
        raise unittest.SkipTest("i6_core not available")
    try:
        import soundfile
    except ImportError:
        raise unittest.SkipTest("soundfile not available")
    rasr_feature_extractor_bin_path = (
        "/work/tools22/asr/rasr/rasr_onnx_haswell_0623/arch/linux-x86_64-standard/"
        "feature-extraction.linux-x86_64-standard"
    )
    if not os.path.exists(rasr_feature_extractor_bin_path):
        raise unittest.SkipTest("RASR feature-extraction binary not found")

    torch.manual_seed(42)
    wav_file_path = tempfile.mktemp(suffix=".wav", prefix="tmp-i6models-random-audio")
    atexit.register(os.remove, wav_file_path)
    generate_random_speech_like_audio_wav(wav_file_path)
    rasr_feature_cache_path = generate_rasr_feature_cache_from_wav_and_flow(
        rasr_feature_extractor_bin_path,
        wav_file_path,
        textwrap.dedent(
            f"""\
            <node filter="generic-vector-s16-demultiplex" name="demultiplex" track="$(track)"/>
            <link from="samples" to="demultiplex"/>
            <node filter="generic-convert-vector-s16-to-vector-f32" name="convert"/>
            <link from="demultiplex" to="convert"/>
            """
        ),
        flow_output_name="convert",
    )

    rasr_cache = FileArchive(rasr_feature_cache_path, must_exists=True)
    time_, rasr_feat = rasr_cache.read("corpus/recording/1", "feat")
    assert len(time_) == len(rasr_feat)
    rasr_feat = torch.tensor(np.concatenate(rasr_feat, axis=0), dtype=torch.float32)
    print("RASR:", _torch_repr(rasr_feat))

    audio, sample_rate = soundfile.read(open(wav_file_path, "rb"), dtype="int16")
    audio = torch.tensor(audio.astype(np.float32))  # [-2**15, 2**15-1]
    print("raw audio", _torch_repr(audio))

    torch.testing.assert_close(audio, rasr_feat, rtol=1e-30, atol=1e-30)


def test_rasr_compatible_preemphasis():
    try:
        from i6_core.lib.rasr_cache import FileArchive
    except ImportError:
        raise unittest.SkipTest("i6_core not available")
    try:
        import soundfile
    except ImportError:
        raise unittest.SkipTest("soundfile not available")
    rasr_feature_extractor_bin_path = (
        "/work/tools22/asr/rasr/rasr_onnx_haswell_0623/arch/linux-x86_64-standard/"
        "feature-extraction.linux-x86_64-standard"
    )
    if not os.path.exists(rasr_feature_extractor_bin_path):
        raise unittest.SkipTest("RASR feature-extraction binary not found")

    torch.manual_seed(42)
    wav_file_path = tempfile.mktemp(suffix=".wav", prefix="tmp-i6models-random-audio")
    atexit.register(os.remove, wav_file_path)
    generate_random_speech_like_audio_wav(wav_file_path)
    rasr_feature_cache_path = generate_rasr_feature_cache_from_wav_and_flow(
        rasr_feature_extractor_bin_path,
        wav_file_path,
        textwrap.dedent(
            f"""\
            <node filter="generic-vector-s16-demultiplex" name="demultiplex" track="$(track)"/>
            <link from="samples" to="demultiplex"/>
            <node filter="generic-convert-vector-s16-to-vector-f32" name="convert"/>
            <link from="demultiplex" to="convert"/>
            <node alpha="1.0" filter="signal-preemphasis" name="preemphasis"/>
            <link from="convert" to="preemphasis"/>
            """
        ),
        flow_output_name="preemphasis",
    )

    rasr_cache = FileArchive(rasr_feature_cache_path, must_exists=True)
    time_, rasr_feat = rasr_cache.read("corpus/recording/1", "feat")
    assert len(time_) == len(rasr_feat)
    rasr_feat = torch.tensor(np.concatenate(rasr_feat, axis=0), dtype=torch.float32)
    print("RASR:", _torch_repr(rasr_feat))

    audio, sample_rate = soundfile.read(open(wav_file_path, "rb"), dtype="int16")
    audio = torch.tensor(audio.astype(np.float32))  # [-2**15, 2**15-1]
    audio[..., 1:] -= 1.0 * audio[..., :-1]
    audio[..., 0] = 0.0
    print("i6_models", _torch_repr(audio))

    torch.testing.assert_close(audio, rasr_feat, rtol=1e-30, atol=1e-30)


def test_rasr_compatible_window():
    try:
        from i6_core.lib.rasr_cache import FileArchive
    except ImportError:
        raise unittest.SkipTest("i6_core not available")
    try:
        import soundfile
    except ImportError:
        raise unittest.SkipTest("soundfile not available")
    rasr_feature_extractor_bin_path = (
        "/work/tools22/asr/rasr/rasr_onnx_haswell_0623/arch/linux-x86_64-standard/"
        "feature-extraction.linux-x86_64-standard"
    )
    if not os.path.exists(rasr_feature_extractor_bin_path):
        raise unittest.SkipTest("RASR feature-extraction binary not found")

    torch.manual_seed(42)
    wav_file_path = tempfile.mktemp(suffix=".wav", prefix="tmp-i6models-random-audio")
    atexit.register(os.remove, wav_file_path)
    generate_random_speech_like_audio_wav(wav_file_path)
    rasr_feature_cache_path = generate_rasr_feature_cache_from_wav_and_flow(
        rasr_feature_extractor_bin_path,
        wav_file_path,
        textwrap.dedent(
            f"""\
            <node filter="generic-vector-s16-demultiplex" name="demultiplex" track="$(track)"/>
            <link from="samples" to="demultiplex"/>
            <node filter="generic-convert-vector-s16-to-vector-f32" name="convert"/>
            <link from="demultiplex" to="convert"/>
            <node alpha="1.0" filter="signal-preemphasis" name="preemphasis"/>
            <link from="convert" to="preemphasis"/>
            <node filter="signal-window" length="0.025" name="window" shift="0.01" type="hanning"/>
            <link from="preemphasis" to="window"/>
            """
        ),
        flow_output_name="window",
    )

    rasr_cache = FileArchive(rasr_feature_cache_path, must_exists=True)
    time_, rasr_feat = rasr_cache.read("corpus/recording/1", "feat")
    assert len(time_) == len(rasr_feat)
    rasr_last_pad = rasr_feat[0].shape[0] - rasr_feat[-1].shape[0]
    rasr_feat[-1] = np.pad(rasr_feat[-1], (0, rasr_last_pad))
    rasr_feat = torch.tensor(np.stack(rasr_feat, axis=0), dtype=torch.float32)
    print("RASR:", _torch_repr(rasr_feat))

    audio, sample_rate = soundfile.read(open(wav_file_path, "rb"), dtype="int16")
    audio = torch.tensor(audio.astype(np.float32))  # [-2**15, 2**15-1]

    # preemphasize
    audio[..., 1:] -= 1.0 * audio[..., :-1]
    audio[..., 0] = 0.0

    # windowing
    win_size = 0.025
    hop_size = 0.01
    hop_length = int(hop_size * sample_rate)
    win_length = int(win_size * sample_rate)

    def _get_length(audio_len: int) -> int:
        if audio_len == 0:
            return 0
        return max(audio_len - win_length + hop_length - 1, 0) // hop_length + 1

    def _get_length_naive(audio_len: int) -> int:
        n = 0
        for t in range(0, audio_len, hop_length):
            n += 1
            if audio_len <= t + win_length:
                break
        return n

    for t in range(10 * win_length):
        assert _get_length(t) == _get_length_naive(t), f"t={t}, {_get_length(t)} != {_get_length_naive(t)}"

    res_len = _get_length(audio.shape[0])
    assert res_len == len(rasr_feat)

    # https://github.com/rwth-i6/rasr/blob/master/src/Signal/WindowFunction.cc
    # https://pytorch.org/docs/stable/generated/torch.hann_window.html

    # manual
    for i, t in enumerate(range(0, audio.shape[0], hop_length)):
        x = audio[t : t + win_length]
        x = x * torch.hann_window(x.shape[0], periodic=False, dtype=torch.float64).to(torch.float32)
        torch.testing.assert_close(x, rasr_feat[i][: x.shape[0]], rtol=1e-30, atol=1e-30)
        # once end was reached, stop
        if audio.shape[0] <= t + win_length:
            assert win_length - x.shape[0] == rasr_last_pad, f"win {win_length}, cur {x.shape[0]}, pad {rasr_last_pad}"
            break

    last_win_size = audio.shape[0] - (res_len - 1) * hop_length
    last_pad = win_length - last_win_size
    assert last_pad == rasr_last_pad, f"last pad {last_pad}, RASR last pad {rasr_last_pad}"
    padded = torch.nn.functional.pad(audio, (0, last_pad))  # zero pad for the last frame

    windowed = padded.unfold(0, size=win_length, step=hop_length)  # [T', W=win_length]
    assert len(windowed) == res_len
    window = torch.hann_window(win_length, periodic=False, dtype=torch.float64).to(torch.float32)
    smoothed = windowed[:-1] * window[None, :]  # [T'-1, W]

    # The last window might be shorter. Will use a shorter Hanning window then. Need to fix that.
    last_win = torch.hann_window(last_win_size, periodic=False, dtype=torch.float64).to(torch.float32)
    last_win = torch.nn.functional.pad(last_win, (0, last_pad))
    smoothed = torch.cat([smoothed, (windowed[-1] * last_win)[None, :]], dim=0)

    print("i6_models", _torch_repr(smoothed))

    torch.testing.assert_close(smoothed, rasr_feat, rtol=1e-30, atol=1e-30)


def test_rasr_compatible_fft():
    try:
        from i6_core.lib.rasr_cache import FileArchive
    except ImportError:
        raise unittest.SkipTest("i6_core not available")
    try:
        import soundfile
    except ImportError:
        raise unittest.SkipTest("soundfile not available")
    rasr_feature_extractor_bin_path = (
        "/work/tools22/asr/rasr/rasr_onnx_haswell_0623/arch/linux-x86_64-standard/"
        "feature-extraction.linux-x86_64-standard"
    )
    if not os.path.exists(rasr_feature_extractor_bin_path):
        raise unittest.SkipTest("RASR feature-extraction binary not found")

    torch.manual_seed(42)
    wav_file_path = tempfile.mktemp(suffix=".wav", prefix="tmp-i6models-random-audio")
    atexit.register(os.remove, wav_file_path)
    generate_random_speech_like_audio_wav(wav_file_path)
    rasr_feature_cache_path = generate_rasr_feature_cache_from_wav_and_flow(
        rasr_feature_extractor_bin_path,
        wav_file_path,
        textwrap.dedent(
            f"""\
            <node filter="generic-vector-s16-demultiplex" name="demultiplex" track="$(track)"/>
            <link from="samples" to="demultiplex"/>
            <node filter="generic-convert-vector-s16-to-vector-f32" name="convert"/>
            <link from="demultiplex" to="convert"/>
            <node alpha="1.0" filter="signal-preemphasis" name="preemphasis"/>
            <link from="convert" to="preemphasis"/>
            <node filter="signal-window" length="0.025" name="window" shift="0.01" type="hanning"/>
            <link from="preemphasis" to="window"/>
            <node filter="signal-real-fast-fourier-transform" maximum-input-size="0.025" name="fft"/>
            <link from="window" to="fft"/>
            <node filter="generic-vector-f32-multiplication" name="scaling" value="16000"/>
            <link from="fft" to="scaling"/>
            """
        ),
        flow_output_name="scaling",
    )

    rasr_cache = FileArchive(rasr_feature_cache_path, must_exists=True)
    time_, rasr_feat = rasr_cache.read("corpus/recording/1", "feat")
    assert len(time_) == len(rasr_feat)
    rasr_feat = torch.tensor(np.stack(rasr_feat, axis=0), dtype=torch.float32)
    print("RASR:", _torch_repr(rasr_feat))

    audio, sample_rate = soundfile.read(open(wav_file_path, "rb"), dtype="int16")
    audio = torch.tensor(audio.astype(np.float32))  # [-2**15, 2**15-1]

    # preemphasize
    audio[..., 1:] -= 1.0 * audio[..., :-1]
    audio[..., 0] = 0.0

    # windowing
    win_size = 0.025
    hop_size = 0.01
    hop_length = int(hop_size * sample_rate)
    win_length = int(win_size * sample_rate)

    res_len = max(audio.shape[0] - win_length + hop_length - 1, 0) // hop_length + 1
    assert res_len == len(rasr_feat)

    last_win_size = audio.shape[0] - (res_len - 1) * hop_length
    last_pad = win_length - last_win_size
    padded = torch.nn.functional.pad(audio, (0, last_pad))  # zero pad for the last frame

    windowed = padded.unfold(0, size=win_length, step=hop_length)  # [T', W=win_length]
    assert len(windowed) == res_len
    window = torch.hann_window(win_length, periodic=False, dtype=torch.float64).to(torch.float32)
    smoothed = windowed[:-1] * window[None, :]  # [T'-1, W]

    # The last window might be shorter. Will use a shorter Hanning window then. Need to fix that.
    last_win = torch.hann_window(last_win_size, periodic=False, dtype=torch.float64).to(torch.float32)
    last_win = torch.nn.functional.pad(last_win, (0, last_pad))
    smoothed = torch.cat([smoothed, (windowed[-1] * last_win)[None, :]], dim=0)

    n_fft = 2 ** math.ceil(math.log2(win_length))
    print(f"win_length={win_length}, n_fft={n_fft}")

    # fft = torch.fft.rfftn(smoothed, s=n_fft)  # [B, T', F]
    # fft = torch.view_as_real(fft).flatten(-2)  # [B, T', F*2]
    # fft = torch.fft.rfftn(smoothed, s=n_fft)  # [B, T', F=n_fft//2+1]
    # fft = torch.view_as_real(fft).flatten(-2)  # [B, T', F=(n_fft//2+1)*2]
    fft = my_fft(smoothed, n_fft=n_fft)

    print("i6_models", _torch_repr(fft))

    torch.testing.assert_close(fft, rasr_feat, rtol=1e-6, atol=1e-2)  # values are huge, accept some errors...?


def test_rasr_compatible_amplitude_spectrum():
    try:
        from i6_core.lib.rasr_cache import FileArchive
    except ImportError:
        raise unittest.SkipTest("i6_core not available")
    try:
        import soundfile
    except ImportError:
        raise unittest.SkipTest("soundfile not available")
    rasr_feature_extractor_bin_path = (
        "/work/tools22/asr/rasr/rasr_onnx_haswell_0623/arch/linux-x86_64-standard/"
        "feature-extraction.linux-x86_64-standard"
    )
    if not os.path.exists(rasr_feature_extractor_bin_path):
        raise unittest.SkipTest("RASR feature-extraction binary not found")

    torch.manual_seed(42)
    wav_file_path = tempfile.mktemp(suffix=".wav", prefix="tmp-i6models-random-audio")
    atexit.register(os.remove, wav_file_path)
    generate_random_speech_like_audio_wav(wav_file_path)
    rasr_feature_cache_path = generate_rasr_feature_cache_from_wav_and_flow(
        rasr_feature_extractor_bin_path,
        wav_file_path,
        textwrap.dedent(
            f"""\
            <node filter="generic-vector-s16-demultiplex" name="demultiplex" track="$(track)"/>
            <link from="samples" to="demultiplex"/>
            <node filter="generic-convert-vector-s16-to-vector-f32" name="convert"/>
            <link from="demultiplex" to="convert"/>
            <node alpha="1.0" filter="signal-preemphasis" name="preemphasis"/>
            <link from="convert" to="preemphasis"/>
            <node filter="signal-window" length="0.025" name="window" shift="0.01" type="hanning"/>
            <link from="preemphasis" to="window"/>
            <node filter="signal-real-fast-fourier-transform" maximum-input-size="0.025" name="fft"/>
            <link from="window" to="fft"/>
            <node filter="generic-vector-f32-multiplication" name="scaling" value="16000"/>
            <link from="fft" to="scaling"/>
            <node filter="signal-vector-alternating-complex-f32-amplitude" name="amplitude-spectrum"/>
            <link from="scaling" to="amplitude-spectrum"/>
            """
        ),
        flow_output_name="amplitude-spectrum",
    )

    rasr_cache = FileArchive(rasr_feature_cache_path, must_exists=True)
    time_, rasr_feat = rasr_cache.read("corpus/recording/1", "feat")
    assert len(time_) == len(rasr_feat)
    rasr_feat = torch.tensor(np.stack(rasr_feat, axis=0), dtype=torch.float32)
    print("RASR:", _torch_repr(rasr_feat))

    audio, sample_rate = soundfile.read(open(wav_file_path, "rb"), dtype="int16")
    audio = torch.tensor(audio.astype(np.float32))  # [-2**15, 2**15-1]

    # preemphasize
    audio[..., 1:] -= 1.0 * audio[..., :-1]
    audio[..., 0] = 0.0

    # windowing
    win_size = 0.025
    hop_size = 0.01
    hop_length = int(hop_size * sample_rate)
    win_length = int(win_size * sample_rate)

    res_len = max(audio.shape[0] - win_length + hop_length - 1, 0) // hop_length + 1
    assert res_len == len(rasr_feat)

    last_win_size = audio.shape[0] - (res_len - 1) * hop_length
    last_pad = win_length - last_win_size
    padded = torch.nn.functional.pad(audio, (0, last_pad))  # zero pad for the last frame

    windowed = padded.unfold(0, size=win_length, step=hop_length)  # [T', W=win_length]
    assert len(windowed) == res_len
    window = torch.hann_window(win_length, periodic=False, dtype=torch.float64).to(torch.float32)
    smoothed = windowed[:-1] * window[None, :]  # [T'-1, W]

    # The last window might be shorter. Will use a shorter Hanning window then. Need to fix that.
    last_win = torch.hann_window(last_win_size, periodic=False, dtype=torch.float64).to(torch.float32)
    last_win = torch.nn.functional.pad(last_win, (0, last_pad))
    smoothed = torch.cat([smoothed, (windowed[-1] * last_win)[None, :]], dim=0)

    n_fft = 2 ** math.ceil(math.log2(win_length))
    print(f"win_length={win_length}, n_fft={n_fft}")

    # fft = torch.fft.rfftn(smoothed, s=n_fft)  # [B, T', F]
    # fft = torch.view_as_real(fft).flatten(-2)  # [B, T', F*2]
    # fft = torch.fft.rfftn(smoothed, s=n_fft)  # [B, T', F=n_fft//2+1]
    # fft = torch.view_as_real(fft).flatten(-2)  # [B, T', F=(n_fft//2+1)*2]
    fft = my_fft(smoothed, n_fft=n_fft)
    amplitude_spectrum = torch.abs(torch.view_as_complex(fft.unflatten(-1, (-1, 2))))

    print("i6_models", _torch_repr(amplitude_spectrum))

    torch.testing.assert_close(amplitude_spectrum, rasr_feat, rtol=1e-30, atol=1e-30)


def create_bit_reversal_reordering(size):
    """
    Creates a bit reversal reordering tensor for a given size.
    """
    # Initial setup
    length = size // 2
    reordering = torch.arange(size)

    # Bit reversal reordering logic
    j = 1
    for i in range(1, size, 2):
        if j > i:
            reordering[i - 1] = j - 1
            reordering[i] = j
        m = length
        while 2 <= m < j:
            j -= m
            m //= 2
        j += m

    return reordering


def bit_reversal_reordering(tensor: torch.Tensor) -> torch.Tensor:
    """
    Reorders a given tensor according to bit reversal pattern.
    """
    size = tensor.size(-1)
    reordering = create_bit_reversal_reordering(size)
    v = tensor.clone()
    for i in range(size):
        h = v[..., i].clone()
        v[..., i] = v[..., reordering[i]]
        v[..., reordering[i]] = h
    return v


def my_fft(tensor: torch.Tensor, *, n_fft: int) -> torch.Tensor:
    # https://github.com/rwth-i6/rasr/blob/master/src/Math/FastFourierTransform.cc#L95
    size = n_fft
    d_pi = torch.tensor(6.28318530717959, dtype=torch.float64)
    theta_base = d_pi

    tensor = torch.nn.functional.pad(tensor, (0, size - tensor.shape[-1]))
    v = bit_reversal_reordering(tensor)

    cur_length = 2
    # estimate DFFT using Danielson and Lanczos formula
    while cur_length < size:
        # initialization of trigonometric recurrence
        step = cur_length * 2
        theta = theta_base / cur_length
        sin_h_theta = torch.sin(0.5 * theta)
        wp_r = -2.0 * sin_h_theta * sin_h_theta
        wp_i = torch.sin(theta)
        w_r = torch.tensor(1.0, dtype=torch.float64)
        w_i = torch.tensor(0.0, dtype=torch.float64)
        for m in range(1, cur_length, 2):
            for i in range(m, size, step):
                # Danielson & Lanczos formula
                j = i + cur_length
                tempr = w_r * v[..., j - 1].to(torch.float64) - w_i * v[..., j].to(torch.float64)
                tempi = w_r * v[..., j].to(torch.float64) + w_i * v[..., j - 1].to(torch.float64)
                tempr = tempr.to(torch.float32)
                tempi = tempi.to(torch.float32)
                v[..., j - 1] = v[..., i - 1] - tempr
                v[..., j] = v[..., i] - tempi
                v[..., i - 1] += tempr
                v[..., i] += tempi
            w_temp_r, w_temp_i = w_r, w_i
            w_r = w_r * wp_r - w_temp_i * wp_i + w_r
            w_i = w_i * wp_r + w_temp_r * wp_i + w_i
        cur_length = step

    pi = torch.tensor(3.141592653589793238, dtype=torch.float64)
    size_d4 = size >> 2
    theta = pi / (size >> 1)

    sin_h_theta = torch.sin(0.5 * theta)
    wp_r = -2.0 * sin_h_theta * sin_h_theta
    wp_i = torch.sin(theta)
    w_r = wp_r + 1
    w_i = wp_i

    for i in range(1, size_d4):
        i1 = i + i
        i2 = i1 + 1
        i3 = size - i1
        i4 = i3 + 1

        # separate the two transforms
        h1_r = 0.5 * (v[..., i1] + v[..., i3]).to(torch.float64)
        h1_i = 0.5 * (v[..., i2] - v[..., i4]).to(torch.float64)
        h2_r = 0.5 * (v[..., i2] + v[..., i4]).to(torch.float64)
        h2_i = -0.5 * (v[..., i1] - v[..., i3]).to(torch.float64)

        # Calculating the true transform of the original real data
        v[..., i1] = h1_r + w_r * h2_r - w_i * h2_i
        v[..., i2] = h1_i + w_r * h2_i + w_i * h2_r
        v[..., i3] = h1_r - w_r * h2_r + w_i * h2_i
        v[..., i4] = -h1_i + w_r * h2_i + w_i * h2_r

        # Updating the trigonometric recurrences
        old_wr = w_r
        w_r = w_r * wp_r - w_i * wp_i + w_r
        w_i = w_i * wp_r + old_wr * wp_i + w_i

    h = v[..., 0].clone()
    v[..., 0] = h + v[..., 1]
    v[..., 1] = h - v[..., 1]

    # unpack logic
    v = torch.nn.functional.pad(v, (0, 2))
    v[..., -2] = v[..., 1]
    v[..., 1] = 0

    return v


def generate_rasr_feature_cache_from_wav_and_flow(
    rasr_feature_extractor_bin_path: str,
    wav_file_path: str,
    flow_network_str: str,
    *,
    flow_input_name: str = "samples",
    flow_output_name: str = "nonlinear",
) -> str:
    import subprocess

    corpus_xml_path = tempfile.mktemp(suffix=".xml", prefix="tmp-rasr-corpus")
    atexit.register(os.remove, corpus_xml_path)
    with open(corpus_xml_path, "w") as f:
        f.write(
            textwrap.dedent(
                f"""\
                <?xml version="1.0" encoding="utf8"?>
                <corpus name="corpus">
                  <recording name="recording" audio="{wav_file_path}">
                    <segment name="1" start="0.0000" end="{_get_wav_file_duration_sec(wav_file_path)}" track="0">
                    </segment>
                  </recording>
                </corpus>
                """
            )
        )

    rasr_feature_cache_path = tempfile.mktemp(suffix=".cache", prefix="tmp-rasr-features")
    atexit.register(os.remove, rasr_feature_cache_path)
    rasr_flow_xml_path = tempfile.mktemp(suffix=".config", prefix="tmp-rasr-flow")
    atexit.register(os.remove, rasr_flow_xml_path)
    with open(rasr_flow_xml_path, "w") as f:
        f.write(
            textwrap.dedent(
                f"""\
                <?xml version="1.0" ?>
                <network name="network">
                  <out name="features"/>
                  <param name="end-time"/>
                  <param name="input-file"/>
                  <param name="start-time"/>
                  <param name="track"/>
                  <param name="id"/>
                  <node filter="audio-input-file-wav" file="$(input-file)"
                   start-time="$(start-time)" end-time="$(end-time)" name="{flow_input_name}"/>
                """
            )
            + textwrap.indent(flow_network_str, "  ")
            + textwrap.dedent(
                f"""\
                  <node filter="generic-cache" id="$(id)" name="cache" path="{rasr_feature_cache_path}"/>
                  <link from="{flow_output_name}" to="cache"/>
                  <link from="cache" to="network:features"/>
                </network>
                """
            )
        )

    rasr_config_path = tempfile.mktemp(suffix=".config", prefix="tmp-rasr-feature-extract")
    atexit.register(os.remove, rasr_config_path)
    with open(rasr_config_path, "w") as f:
        f.write(
            textwrap.dedent(
                f"""\
                [*.corpus]
                file = {corpus_xml_path}

                [*.feature-extraction]
                file = {rasr_flow_xml_path}
                """
            )
        )

    subprocess.check_call([rasr_feature_extractor_bin_path, "--config", rasr_config_path])
    return rasr_feature_cache_path


def _get_wav_file_duration_sec(wav_file_path: str) -> float:
    import wave
    import contextlib

    with contextlib.closing(wave.open(wav_file_path, "r")) as f:
        frames = f.getnframes()
        rate = f.getframerate()
        return frames / float(rate)


def generate_random_speech_like_audio_wav(
    output_wav_file_path: str,
    duration_sec: float = 5.0,
    *,
    samples_per_sec: int = 16_000,
    sample_width_bytes: int = 2,  # int16
    frequency: float = 150.0,
    num_random_freqs_per_sec: int = 15,
    amplitude: float = 0.3,
    amplitude_frequency: Optional[float] = None,
):
    import wave

    f = wave.open(output_wav_file_path, "wb")
    f.setframerate(samples_per_sec)
    f.setnchannels(1)
    f.setsampwidth(sample_width_bytes)

    samples = generate_random_speech_like_audio(
        batch_size=1,
        num_frames=int(duration_sec * samples_per_sec),
        samples_per_sec=samples_per_sec,
        frequency=frequency,
        num_random_freqs_per_sec=num_random_freqs_per_sec,
        amplitude=amplitude,
        amplitude_frequency=amplitude_frequency,
    )  # [B,T]
    samples = samples[0]  # [T]
    print("generated raw samples:", _torch_repr(samples))

    samples_int = (samples * (2 ** (8 * sample_width_bytes - 1) - 1)).to(
        {1: torch.int8, 2: torch.int16, 4: torch.int32}[sample_width_bytes]
    )

    f.writeframes(samples_int.numpy().tobytes())
    f.close()


def generate_random_speech_like_audio(
    batch_size: int,
    num_frames: int,
    *,
    samples_per_sec: int = 16_000,
    frequency: float = 150.0,
    num_random_freqs_per_sec: int = 15,
    amplitude: float = 0.3,
    amplitude_frequency: Optional[float] = None,
) -> torch.Tensor:
    """
    generate audio

    Source:
    https://github.com/albertz/playground/blob/master/create-random-speech-like-sound.py

    :return: shape [batch_size,num_frames]
    """
    frame_idxs = torch.arange(num_frames, dtype=torch.int64)  # [T]

    samples = _integrate_rnd_frequencies(
        batch_size,
        frame_idxs,
        base_frequency=frequency,
        samples_per_sec=samples_per_sec,
        num_random_freqs_per_sec=num_random_freqs_per_sec,
    )  # [T,B]

    if amplitude_frequency is None:
        amplitude_frequency = frequency / 75.0
    amplitude_variations = _integrate_rnd_frequencies(
        batch_size,
        frame_idxs,
        base_frequency=amplitude_frequency,
        samples_per_sec=samples_per_sec,
        num_random_freqs_per_sec=amplitude_frequency,
    )  # [T,B]

    samples *= amplitude * (0.666 + 0.333 * amplitude_variations)
    return samples.permute(1, 0)  # [B,T]


def _integrate_rnd_frequencies(
    batch_size: int,
    frame_idxs: torch.Tensor,
    *,
    base_frequency: float,
    samples_per_sec: int,
    num_random_freqs_per_sec: float,
) -> torch.Tensor:
    rnd_freqs = torch.empty(
        size=(int(len(frame_idxs) * num_random_freqs_per_sec / samples_per_sec) + 1, batch_size),
        dtype=torch.float32,
    )  # [T',B]
    torch.nn.init.trunc_normal_(rnd_freqs, a=-1.0, b=1.0)
    rnd_freqs = (rnd_freqs * 0.5 + 1.0) * base_frequency  # [T',B]

    freq_idx_f = (frame_idxs * num_random_freqs_per_sec) / samples_per_sec
    freq_idx = freq_idx_f.to(torch.int64)
    next_freq_idx = torch.clip(freq_idx + 1, 0, len(rnd_freqs) - 1)
    frac = (freq_idx_f % 1)[:, None]  # [T,1]
    freq = rnd_freqs[freq_idx] * (1 - frac) + rnd_freqs[next_freq_idx] * frac  # [T,B]

    ts = torch.cumsum(freq / samples_per_sec, dim=0)  # [T,B]
    return torch.sin(2 * torch.pi * ts)


def _torch_repr(x: torch.Tensor) -> str:
    try:
        from lovely_tensors import lovely
    except ImportError:
        mean, std = x.mean(), x.std()
        min_, max_ = x.min(), x.max()
        return f"{x.shape} x∈[{min_}, {max_}] μ={mean} σ={std} {x.dtype}"
    else:
        return lovely(x)


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(f"*** {arg}()")
        globals()[arg]()



from dataclasses import dataclass
import torch
from torch import nn

from i6_models.config import ModelConfiguration


@dataclass
class BlstmEncoderV1Config(ModelConfiguration):
    """
    Attributes:
        num_layers: number of bi-directional LSTM layers, minimum 2
        input_dim: input dimension size
        hidden_dim: hidden dimension of one direction of LSTM, the total output size is twice of this
        dropout: nn.LSTM supports internal Dropout applied between each layer of BLSTM (but not on input/output)
        enforce_sorted:
            True: expects that sequences are sorted by sequence length in decreasing order.
                Will not do any sorting.
                This is required for ONNX-Export, and thus the recommended setting.
            False: no expectation.
                It will internally enforce that they are sorted
                and undo the reordering at the output.

            Sorting can for example be performed independent of the ONNX export in e.g. train_step:

                audio_features_len, indices = torch.sort(audio_features_len, descending=True)
                audio_features = audio_features[indices, :, :]
                labels = labels[indices, :]
                labels_len = labels_len[indices]
    """

    num_layers: int
    input_dim: int
    hidden_dim: int
    dropout: float
    enforce_sorted: bool


class BlstmEncoderV1(torch.nn.Module):
    """
    Simple multi-layer BLSTM model including dropout, batch-first variant,
    hardcoded to use B,T,F input

    supports: TorchScript, ONNX-export
    """

    def __init__(self, config: BlstmEncoderV1Config):
        """
        :param config: configuration object
        """
        super().__init__()
        self.dropout = config.dropout
        self.enforce_sorted = config.enforce_sorted
        self.blstm_stack = nn.LSTM(
            input_size=config.input_dim,
            hidden_size=config.hidden_dim,
            bidirectional=True,
            num_layers=config.num_layers,
            batch_first=True,
            dropout=self.dropout,
        )

    def forward(self, x: torch.Tensor, seq_len: torch.Tensor) -> torch.Tensor:
        """
        :param x: [B, T, input_dim]
        :param seq_len: [B], should be on CPU for Script/Trace mode
        :return [B, T, 2 * hidden_dim]
        """
        if not torch.jit.is_scripting() and not torch.jit.is_tracing():
            # during graph mode we have to assume all Tensors are on the correct device,
            # otherwise move lengths to the CPU if they are on GPU
            if seq_len.get_device() >= 0:
                seq_len = seq_len.cpu()

        blstm_packed_in = nn.utils.rnn.pack_padded_sequence(
            input=x,
            lengths=seq_len,
            enforce_sorted=self.enforce_sorted,
            batch_first=True,
        )
        blstm_out, _ = self.blstm_stack(blstm_packed_in)
        blstm_out, _ = nn.utils.rnn.pad_packed_sequence(blstm_out, padding_value=0.0, batch_first=True)

        return blstm_out



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Tuple, Sequence, Union, Callable, List

import torch
from torch import nn

from i6_models.parts.frontend.generic_frontend import FrontendLayerType, GenericFrontendV1, GenericFrontendV1Config


@dataclass
class GenericFrontendV1TestParams:
    batch: int
    time: int
    in_features: int
    layer_ordering: Sequence[FrontendLayerType]
    conv_kernel_sizes: Optional[Sequence[Tuple[int, int]]]
    conv_strides: Optional[Sequence[Tuple[int, int]]]
    conv_paddings: Optional[Sequence[Tuple[int, int]]]
    conv_out_dims: Optional[Sequence[int]]
    pool_kernel_sizes: Optional[Sequence[Tuple[int, int]]]
    pool_strides: Optional[Sequence[Tuple[int, int]]]
    pool_paddings: Optional[Sequence[Tuple[int, int]]]
    activations: Optional[Sequence[Union[nn.Module, Callable[[torch.Tensor], torch.Tensor]]]]
    out_features: int
    output_shape: List[int]
    in_sequence_mask: torch.Tensor
    out_sequence_mask: torch.Tensor


def test_generic_frontend_v1():
    torch.manual_seed(42)

    def get_output_shape(test_parameters: GenericFrontendV1TestParams):
        data_input = torch.randn(
            test_parameters.batch,
            test_parameters.time,
            test_parameters.in_features,
        )

        cfg = GenericFrontendV1Config(
            in_features=test_parameters.in_features,
            layer_ordering=test_parameters.layer_ordering,
            conv_kernel_sizes=test_parameters.conv_kernel_sizes,
            conv_strides=test_parameters.conv_strides,
            conv_paddings=test_parameters.conv_paddings,
            conv_out_dims=test_parameters.conv_out_dims,
            pool_kernel_sizes=test_parameters.pool_kernel_sizes,
            pool_strides=test_parameters.pool_strides,
            pool_paddings=test_parameters.pool_paddings,
            activations=test_parameters.activations,
            out_features=test_parameters.out_features,
        )

        output, sequence_mask = GenericFrontendV1(cfg)(
            data_input,
            test_parameters.in_sequence_mask,
        )

        return output.shape, sequence_mask

    for idx, test_params in enumerate(
        [
            GenericFrontendV1TestParams(
                batch=10,
                time=50,
                in_features=50,
                layer_ordering=[FrontendLayerType.Conv2d, FrontendLayerType.Conv2d],
                conv_kernel_sizes=[(3, 3), (5, 5)],
                conv_strides=[(1, 1), (1, 1)],
                conv_paddings=None,
                conv_out_dims=[32, 32],
                pool_kernel_sizes=None,
                pool_strides=None,
                pool_paddings=None,
                activations=None,
                out_features=384,
                output_shape=[10, 50, 384],
                in_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),
                out_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),
            ),
            GenericFrontendV1TestParams(
                batch=10,
                time=50,
                in_features=50,
                layer_ordering=[
                    FrontendLayerType.Conv2d,
                    FrontendLayerType.Pool2d,
                    FrontendLayerType.Conv2d,
                    FrontendLayerType.Activation,
                ],
                conv_kernel_sizes=[(3, 3), (5, 5)],
                conv_strides=[(1, 1), (1, 1)],
                conv_paddings=None,
                conv_out_dims=[32, 32],
                pool_kernel_sizes=[(3, 3)],
                pool_strides=[(2, 2)],
                pool_paddings=None,
                activations=[nn.SiLU()],
                out_features=384,
                output_shape=[10, 25, 384],
                in_sequence_mask=torch.Tensor(10 * [50 * [True] + 0 * [False]]).bool(),
                out_sequence_mask=torch.Tensor(10 * [25 * [True] + 0 * [False]]).bool(),
            ),
            GenericFrontendV1TestParams(
                batch=10,
                time=50,
                in_features=50,
                layer_ordering=[
                    FrontendLayerType.Conv2d,
                    FrontendLayerType.Pool2d,
                    FrontendLayerType.Activation,
                    FrontendLayerType.Conv2d,
                    FrontendLayerType.Activation,
                ],
                conv_kernel_sizes=[(3, 3), (3, 5)],
                conv_strides=[(1, 1), (2, 1)],
                conv_paddings=None,
                conv_out_dims=[32, 32],
                pool_kernel_sizes=[(3, 3)],
                pool_strides=[(2, 2)],
                pool_paddings=None,
                activations=[nn.SiLU(), nn.SiLU()],
                out_features=384,
                output_shape=[10, 13, 384],
                in_sequence_mask=torch.Tensor(10 * [50 * [True] + 0 * [False]]).bool(),
                out_sequence_mask=torch.Tensor(10 * [13 * [True] + 0 * [False]]).bool(),
            ),
            GenericFrontendV1TestParams(
                batch=10,
                time=50,
                in_features=50,
                layer_ordering=[
                    FrontendLayerType.Conv2d,
                    FrontendLayerType.Pool2d,
                    FrontendLayerType.Activation,
                    FrontendLayerType.Conv2d,
                    FrontendLayerType.Pool2d,
                    FrontendLayerType.Activation,
                ],
                conv_kernel_sizes=[(3, 3), (3, 5)],
                conv_strides=[(1, 1), (2, 2)],
                conv_paddings=None,
                conv_out_dims=[32, 32],
                pool_kernel_sizes=[(3, 3), (5, 5)],
                pool_strides=[(1, 2), (2, 3)],
                pool_paddings=None,
                activations=[nn.SiLU(), nn.SiLU()],
                out_features=384,
                output_shape=[10, 13, 384],
                in_sequence_mask=torch.Tensor(10 * [50 * [True] + 0 * [False]]).bool(),
                out_sequence_mask=torch.Tensor(10 * [13 * [True] + 0 * [False]]).bool(),
            ),
        ]
    ):
        shape, seq_mask = get_output_shape(test_params)
        assert list(shape) == test_params.output_shape, (shape, test_params.output_shape)
        assert torch.equal(seq_mask, test_params.out_sequence_mask), (
            seq_mask.shape,
            test_params.out_sequence_mask.shape,
        )



__all__ = [
    "WindowConvolutionFrontendV1Config",
    "WindowConvolutionFrontendV1",
]

from dataclasses import dataclass
from typing import Callable, Optional, Tuple, Union

import torch
from torch import nn
from torch.nn import functional as F

from i6_models.config import ModelConfiguration

from .common import mask_pool, apply_same_padding


@dataclass
class WindowConvolutionFrontendV1Config(ModelConfiguration):
    """
    Attributes:
        input_dim: number of input features to module
        output_dim: output dimension
        dropout: dropout after linear layer
        kernel_size: number of feature frames to convolve
        stride: skip (stride - 1) feature frames; stride > 1 implies subsampling
        activation: activation function applied after linear computation
    """

    input_dim: int
    output_dim: int
    dropout: float
    kernel_size: int
    stride: int
    activation: Union[nn.Module, Callable[[torch.Tensor], torch.Tensor]]

    def __post_init__(self):
        super().__post_init__()
        assert self.stride >= 1, "Choose an integer >= 1 for stride"
        assert 0.0 <= self.dropout <= 1.0, "Dropout value must be a probability"


class WindowConvolutionFrontendV1(nn.Module):
    """
    Simple feed-forward front-end that computes over a window
    of input features. Choosing a stride > 1 allows for subsampling
    of the features.
    """

    def __init__(self, cfg: WindowConvolutionFrontendV1Config):
        """
        :param cfg: model configuration for this module
        """
        super().__init__()
        self.conv = torch.nn.Conv1d(
            in_channels=cfg.input_dim,
            out_channels=cfg.output_dim,
            kernel_size=cfg.kernel_size,
            stride=cfg.stride,
            padding=0,
            bias=True,
        )
        self.activation = cfg.activation
        self.pad = lambda x: apply_same_padding(x, cfg.kernel_size)
        self.dropout = torch.nn.Dropout(cfg.dropout)

    def forward(self, x: torch.Tensor, /, sequence_mask: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        T might be reduced to T' on stride

        :param x: input tensor of shape [B,T,F]
        :param sequence_mask: the sequence mask for the tensor
        :return: torch.Tensor of shape [B,T',F'] and the shape of the sequence mask
        """
        # torch 1d convolution is over last dim but we want time conv
        x = x.transpose(1, 2)  # [B, F, T]
        x = self.pad(x)
        x = self.conv(x).transpose(1, 2)  # [B, T', F']

        # change masking according to stride value
        sequence_mask = mask_pool(
            sequence_mask,
            kernel_size=1,
            stride=self.conv.stride[0],
            padding=0,  # done manually
        )
        x = self.activation(x)
        x = self.dropout(x)

        return x, sequence_mask



from dataclasses import dataclass
import pytest

from i6_models.config import ModelConfiguration, ModuleFactoryV1
from torch import nn


def test_simple_configuration():
    @dataclass
    class TestConfiguration(ModelConfiguration):
        num_layers: int = 5
        hidden_dim: int = 256
        name: str = "Cool Model Configuration"

    test_cfg = TestConfiguration(num_layers=12, name="Even Cooler Model Configuration")
    assert test_cfg.hidden_dim == 256
    assert test_cfg.name == "Even Cooler Model Configuration"
    assert test_cfg.num_layers == 12
    test_cfg.num_layers = 7
    assert test_cfg.num_layers == 7


def test_nested_configuration():
    @dataclass
    class TestConfiguration(ModelConfiguration):
        num_layers: int = 5
        hidden_dim: int = 256
        name: str = "Cool Model Configuration"

    @dataclass
    class TestNestedConfiguration(ModelConfiguration):
        encoder_config: TestConfiguration = TestConfiguration(num_layers=4, hidden_dim=3, name="encoder_config")
        decoder_config: TestConfiguration = TestConfiguration(num_layers=6, hidden_dim=5, name="decoder_config")

    dec_cfg = TestConfiguration()
    test_cfg = TestNestedConfiguration(decoder_config=dec_cfg)

    assert test_cfg.encoder_config.num_layers == 4
    assert test_cfg.encoder_config.hidden_dim == 3
    assert test_cfg.encoder_config.name == "encoder_config"
    assert test_cfg.decoder_config.num_layers == 5
    assert test_cfg.decoder_config.hidden_dim == 256
    assert test_cfg.decoder_config.name == "Cool Model Configuration"
    test_cfg.encoder_config = TestConfiguration(num_layers=1, hidden_dim=2, name="better_encoder_config")
    assert test_cfg.encoder_config.num_layers == 1
    assert test_cfg.encoder_config.hidden_dim == 2
    assert test_cfg.encoder_config.name == "better_encoder_config"


def test_config_typing():
    @dataclass
    class TestConfiguration(ModelConfiguration):
        num_layers: int = 4
        hidden_dim: int = 13
        name: str = "Cool Model Configuration"

    from typeguard import TypeCheckError

    TestConfiguration(num_layers=2, hidden_dim=1)
    with pytest.raises(TypeCheckError):
        TestConfiguration(num_layers=2.0, hidden_dim="One")


def test_module_factory():
    @dataclass
    class TestConfiguration(ModelConfiguration):
        param: int = 4

    @dataclass
    class TestConfiguration2(ModelConfiguration):
        param: float = 4.3

    class TestModule(nn.Module):
        def __init__(self, cfg: TestConfiguration):
            super().__init__()
            self.cfg = cfg

        def forward(self):
            pass

    factory = ModuleFactoryV1(module_class=TestModule, cfg=TestConfiguration())
    obj = factory()
    assert type(obj) == TestModule
    obj2 = factory()
    assert obj != obj2
    from typeguard import TypeCheckError

    with pytest.raises(TypeCheckError):
        ModuleFactoryV1(module_class=TestModule, cfg=TestConfiguration2())



__all__ = ["LayerNormNC"]
import torch
import torch.nn as nn


class LayerNormNC(nn.LayerNorm):
    """
    LayerNorm that accepts [N,C,*] tensors and normalizes over C (channels) dimension.
    see here: https://pytorch.org/docs/stable/generated/torch.nn.LayerNorm.html
    """

    def __init__(self, channels: int):
        """
        :param channels: number of channels for normalization
        """
        super().__init__(channels)

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        :param tensor: input tensor with shape [N,C,*]
        :return: normalized tensor with shape [N,C,*]
        """
        return super().forward(tensor.transpose(1, -1)).transpose(1, -1)



from __future__ import annotations

__all__ = ["ConformerMHSAV1", "ConformerMHSAV1Config"]
from dataclasses import dataclass
import torch

from i6_models.config import ModelConfiguration
from i6_models.util import compat


@dataclass
class ConformerMHSAV1Config(ModelConfiguration):
    """
    Attributes:
        input_dim: input dim and total dimension for query/key and value projections, should be divisible by `num_att_heads`
        num_att_heads: number of attention heads
        att_weights_dropout: attention weights dropout
        dropout: multi-headed self attention output dropout
    """

    input_dim: int
    num_att_heads: int
    att_weights_dropout: float
    dropout: float

    def __post_init__(self) -> None:
        super().__post_init__()
        assert self.input_dim % self.num_att_heads == 0, "input_dim must be divisible by num_att_heads"


class ConformerMHSAV1(torch.nn.Module):
    """
    Conformer multi-headed self-attention module
    """

    def __init__(self, cfg: ConformerMHSAV1Config):

        super().__init__()

        self.layernorm = torch.nn.LayerNorm(cfg.input_dim)
        self.mhsa = torch.nn.MultiheadAttention(
            cfg.input_dim, cfg.num_att_heads, dropout=cfg.att_weights_dropout, batch_first=True
        )
        self.dropout = cfg.dropout

    def forward(self, input_tensor: torch.Tensor, sequence_mask: torch.Tensor) -> torch.Tensor:
        """
        Apply layer norm and multi-head self attention and dropout

        :param input_tensor: Input to the self attention of shape (B, T, F)
        :param sequence_mask: bool mask of shape (B, T), True signals within sequence, False outside, will be inverted to match the torch.nn.MultiheadAttention module
        which will be applied/added to dot product, used to mask padded key positions out
        """
        inv_sequence_mask = compat.logical_not(sequence_mask)
        output_tensor = self.layernorm(input_tensor)  # [B,T,F]

        output_tensor, _ = self.mhsa(
            output_tensor, output_tensor, output_tensor, key_padding_mask=inv_sequence_mask, need_weights=False
        )  # [B,T,F]
        output_tensor = torch.nn.functional.dropout(output_tensor, p=self.dropout, training=self.training)  # [B,T,F]

        return output_tensor



"""
Compatibility support for different functions. This could be for example for onnx export.
"""

import torch


def logical_not(tensor: torch.Tensor, /) -> torch.Tensor:
    """
    Helper function to decide how to invert the sequence mask. For ONNX export use XOR with 1 since logical_not is not implemented.
    Else logical_not is applied for efficiency reasons.

    :param tensor: bool mask of shape (B, T) to be inverted.
    """
    if torch.onnx.is_in_onnx_export():
        return torch.logical_xor(tensor, torch.ones_like(tensor))
    else:
        return torch.logical_not(tensor)



from __future__ import annotations

__all__ = ["ConformerBlockV1Config", "ConformerEncoderV1Config", "ConformerBlockV1", "ConformerEncoderV1"]

import torch
from torch import nn
from dataclasses import dataclass
from typing import Tuple

from i6_models.config import ModelConfiguration, ModuleFactoryV1
from i6_models.parts.conformer import (
    ConformerConvolutionV1,
    ConformerConvolutionV1Config,
    ConformerMHSAV1,
    ConformerMHSAV1Config,
    ConformerPositionwiseFeedForwardV1,
    ConformerPositionwiseFeedForwardV1Config,
)


@dataclass
class ConformerBlockV1Config(ModelConfiguration):
    """
    Attributes:
        ff_cfg: Configuration for ConformerPositionwiseFeedForwardV1
        mhsa_cfg: Configuration for ConformerMHSAV1
        conv_cfg: Configuration for ConformerConvolutionV1
    """

    # nested configurations
    ff_cfg: ConformerPositionwiseFeedForwardV1Config
    mhsa_cfg: ConformerMHSAV1Config
    conv_cfg: ConformerConvolutionV1Config


class ConformerBlockV1(nn.Module):
    """
    Conformer block module
    """

    def __init__(self, cfg: ConformerBlockV1Config):
        """
        :param cfg: conformer block configuration with subunits for the different conformer parts
        """
        super().__init__()
        self.ff1 = ConformerPositionwiseFeedForwardV1(cfg=cfg.ff_cfg)
        self.mhsa = ConformerMHSAV1(cfg=cfg.mhsa_cfg)
        self.conv = ConformerConvolutionV1(model_cfg=cfg.conv_cfg)
        self.ff2 = ConformerPositionwiseFeedForwardV1(cfg=cfg.ff_cfg)
        self.final_layer_norm = torch.nn.LayerNorm(cfg.ff_cfg.input_dim)

    def forward(self, x: torch.Tensor, /, sequence_mask: torch.Tensor) -> torch.Tensor:
        """
        :param x: input tensor of shape [B, T, F]
        :param sequence_mask: mask tensor where 1 defines positions within the sequence and 0 outside, shape: [B, T]
        :return: torch.Tensor of shape [B, T, F]
        """
        x = 0.5 * self.ff1(x) + x  #  [B, T, F]
        x = self.mhsa(x, sequence_mask) + x  #  [B, T, F]
        x = self.conv(x) + x  #  [B, T, F]
        x = 0.5 * self.ff2(x) + x  #  [B, T, F]
        x = self.final_layer_norm(x)  #  [B, T, F]
        return x


@dataclass
class ConformerEncoderV1Config(ModelConfiguration):
    """
    Attributes:
        num_layers: Number of conformer layers in the conformer encoder
        frontend: A pair of ConformerFrontend and corresponding config
        block_cfg: Configuration for ConformerBlockV1
    """

    num_layers: int

    # nested configurations
    frontend: ModuleFactoryV1
    block_cfg: ConformerBlockV1Config


class ConformerEncoderV1(nn.Module):
    """
    Implementation of the convolution-augmented Transformer (short Conformer), as in the original publication.
    The model consists of a frontend and a stack of N conformer blocks.
    C.f. https://arxiv.org/pdf/2005.08100.pdf
    """

    def __init__(self, cfg: ConformerEncoderV1Config):
        """
        :param cfg: conformer encoder configuration with subunits for frontend and conformer blocks
        """
        super().__init__()

        self.frontend = cfg.frontend()
        self.module_list = torch.nn.ModuleList([ConformerBlockV1(cfg.block_cfg) for _ in range(cfg.num_layers)])

    def forward(self, data_tensor: torch.Tensor, sequence_mask: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        :param data_tensor: input tensor of shape [B, T', F]
        :param sequence_mask: mask tensor where 1 defines positions within the sequence and 0 outside, shape: [B, T']
        :return: (output, out_seq_mask)
            where output is torch.Tensor of shape [B, T, F'],
            out_seq_mask is a torch.Tensor of shape [B, T]

        F: input feature dim, F': internal and output feature dim
        T': data time dim, T: down-sampled time dim (internal time dim)
        """
        x, sequence_mask = self.frontend(data_tensor, sequence_mask)  # [B, T, F']
        for module in self.module_list:
            x = module(x, sequence_mask)  # [B, T, F']

        return x, sequence_mask



from itertools import product

import torch
from torch import nn

from i6_models.parts.e_branchformer.cgmlp import ConvolutionalGatingMLPV1Config, ConvolutionalGatingMLPV1
from i6_models.parts.e_branchformer.merge import MergerV1Config, MergerV1


def test_ConvolutionalGatingMLPV1():
    def get_output_shape(input_shape, hidden_dim, kernel_size, dropout, activation):
        input_dim = input_shape[-1]
        cfg = ConvolutionalGatingMLPV1Config(input_dim, hidden_dim, kernel_size, dropout, activation)
        e_branchformer_cgmlp_part = ConvolutionalGatingMLPV1(cfg)
        x = torch.randn(input_shape)
        y = e_branchformer_cgmlp_part(x)
        return y.shape

    for input_shape, hidden_dim, kernel_size, dropout, activation in product(
        [(100, 5, 20), (200, 30, 10)], [120, 60], [9, 15], [0.1, 0.3], [nn.functional.gelu, nn.functional.relu]
    ):
        assert get_output_shape(input_shape, hidden_dim, kernel_size, dropout, activation) == input_shape


def test_MergerV1():
    def get_output_shape(input_shape, kernel_size, dropout):
        input_dim = input_shape[-1]
        cfg = MergerV1Config(input_dim, kernel_size, dropout)
        e_branchformer_merge_part = MergerV1(cfg)
        tensor_local = torch.randn(input_shape)
        tensor_global = torch.randn(input_shape)
        y = e_branchformer_merge_part(tensor_local, tensor_global)
        return y.shape

    for input_shape, kernel_size, dropout in product([(100, 5, 20), (200, 30, 10)], [15, 31], [0.1, 0.3]):
        assert get_output_shape(input_shape, kernel_size, dropout) == input_shape



"""
i6_models root import
"""

from .__setup__ import get_version_str as _get_version_str

__long_version__ = _get_version_str(
    fallback="1.0.0+unknown", long=True, verbose_error=True
)  # `SemVer <https://semver.org/>`__ compatible
__version__ = __long_version__[: __long_version__.index("+")]  # distutils.version.StrictVersion compatible
__git_version__ = __long_version__  # just an alias, to keep similar to other projects



from __future__ import annotations
from itertools import product

import torch
from torch import nn

from i6_models.parts.conformer.convolution import ConformerConvolutionV1, ConformerConvolutionV1Config
from i6_models.parts.conformer.feedforward import (
    ConformerPositionwiseFeedForwardV1,
    ConformerPositionwiseFeedForwardV1Config,
)
from i6_models.parts.conformer.mhsa import ConformerMHSAV1Config, ConformerMHSAV1
from i6_models.parts.conformer.norm import LayerNormNC


def test_conformer_convolution_output_shape():
    def get_output_shape(batch, time, features, norm=None, kernel_size=31, dropout=0.1, activation=nn.functional.silu):
        x = torch.randn(batch, time, features)
        if norm is None:
            norm = nn.BatchNorm1d(features)
        cfg = ConformerConvolutionV1Config(
            channels=features, kernel_size=kernel_size, dropout=dropout, activation=activation, norm=norm
        )
        conformer_conv_part = ConformerConvolutionV1(cfg)
        y = conformer_conv_part(x)
        return y.shape

    assert get_output_shape(10, 50, 250) == (10, 50, 250)
    assert get_output_shape(10, 50, 250, activation=nn.functional.relu) == (10, 50, 250)  # different activation
    assert get_output_shape(10, 50, 250, norm=LayerNormNC(250)) == (10, 50, 250)  # different norm
    assert get_output_shape(1, 50, 100) == (1, 50, 100)  # test with batch size 1
    assert get_output_shape(10, 1, 50) == (10, 1, 50)  # time dim 1
    assert get_output_shape(10, 10, 20, dropout=0.0) == (10, 10, 20)  # dropout 0
    assert get_output_shape(10, 10, 20, kernel_size=3) == (10, 10, 20)  # odd kernel size


def test_ConformerPositionwiseFeedForwardV1():
    def get_output_shape(input_shape, input_dim, hidden_dim, dropout, activation):
        x = torch.randn(input_shape)
        cfg = ConformerPositionwiseFeedForwardV1Config(input_dim, hidden_dim, dropout, activation)
        conf_ffn_part = ConformerPositionwiseFeedForwardV1(cfg)
        y = conf_ffn_part(x)
        return y.shape

    for input_dim, hidden_dim, dropout, activation in product(
        [10, 20], [100, 200], [0.1, 0.3], [nn.functional.silu, nn.functional.relu]
    ):
        input_shape = (10, 100, input_dim)
        assert get_output_shape(input_shape, input_dim, hidden_dim, dropout, activation) == input_shape


def test_ConformerMHSAV1():
    def get_output_shape(input_shape, cfg, **kwargs):

        input = torch.randn(input_shape)
        output = ConformerMHSAV1(cfg)(input, **kwargs)

        return list(output.shape)

    # with key padding mask
    input_shape = [4, 15, 32]  # B,T,F
    cfg = ConformerMHSAV1Config(32, 8, 0.2, 0.3)
    assert get_output_shape(input_shape, cfg, sequence_mask=(torch.randint(0, 2, input_shape[:2]) > 0)) == [4, 15, 32]


def test_layer_norm_nc():
    torch.manual_seed(42)

    def get_output(shape, norm):
        x = torch.randn(shape)
        out = norm(x)
        return out

    # test with different shape
    torch_ln = get_output([10, 50, 250], nn.LayerNorm(250))
    custom_ln = get_output([10, 250, 50], LayerNormNC(250))
    torch.allclose(torch_ln, custom_ln.transpose(1, 2))

    torch_ln = get_output([10, 8, 23], nn.LayerNorm(23))
    custom_ln = get_output([10, 23, 8], LayerNormNC(23))
    torch.allclose(torch_ln, custom_ln.transpose(1, 2))



from itertools import product

import torch
from torch import nn
from torch.nn import functional as F

from i6_models.parts.frontend.window_convolution import WindowConvolutionFrontendV1Config, WindowConvolutionFrontendV1


def test_output_shape():
    in_features = 80
    out_features = 2048
    dropout = 0.1
    max_seq_lens = 100

    for window_size, stride in product(range(1, 22), range(1, 5)):
        frontend = WindowConvolutionFrontendV1(
            WindowConvolutionFrontendV1Config(
                input_dim=80,
                output_dim=out_features,
                kernel_size=window_size,
                dropout=dropout,
                stride=stride,
                activation=F.relu,
            )
        )

        feat_len = torch.arange(start=1, end=max_seq_lens + 1)
        mask = torch.less(torch.arange(max_seq_lens)[None, :], feat_len[:, None])

        features = torch.empty((max_seq_lens, max_seq_lens, in_features))

        out, out_mask = frontend(features, mask)

        expected_out_len = (feat_len - 1) // stride + 1
        expected_shape = (max_seq_lens, expected_out_len[-1], out_features)
        assert out.shape == expected_shape, f"Output with shape {out.shape} not as expected {expected_shape}"
        for i in range(expected_out_len[-1] - 1):
            # check if masks are correct
            assert (
                out_mask[i, expected_out_len[i] - 1] and not out_mask[i, expected_out_len[i]]
            ), f"Failed for {i=}, {stride=}, {window_size=}, {out_mask[i]=}, {out_mask[i].shape=}"



from __future__ import annotations

__all__ = [
    "EbranchformerBlockV1Config",
    "EbranchformerBlockV1",
    "EbranchformerEncoderV1Config",
    "EbranchformerEncoderV1",
]

import torch
from torch import nn
from dataclasses import dataclass
from typing import Tuple

from i6_models.config import ModelConfiguration, ModuleFactoryV1
from i6_models.parts.conformer import (
    ConformerMHSAV1 as MHSAV1,
    ConformerMHSAV1Config as MHSAV1Config,
    ConformerPositionwiseFeedForwardV1 as PositionwiseFeedForwardV1,
    ConformerPositionwiseFeedForwardV1Config as PositionwiseFeedForwardV1Config,
)
from i6_models.parts.e_branchformer import (
    ConvolutionalGatingMLPV1Config,
    ConvolutionalGatingMLPV1,
    MergerV1Config,
    MergerV1,
)


@dataclass
class EbranchformerBlockV1Config(ModelConfiguration):
    """
    Attributes:
        ff_cfg: Configuration for PositionwiseFeedForwardV1 module
        mhsa_cfg: Configuration for MHSAV1 module
        cgmlp_cfg: Configuration for ConvolutionalGatingMLPV1 module
        merger_cfg: Configuration for MergerV1 module
    """

    ff_cfg: PositionwiseFeedForwardV1Config
    mhsa_cfg: MHSAV1Config
    cgmlp_cfg: ConvolutionalGatingMLPV1Config
    merger_cfg: MergerV1Config


class EbranchformerBlockV1(nn.Module):
    """
    Ebranchformer block module
    """

    def __init__(self, cfg: EbranchformerBlockV1Config):
        """
        :param cfg: e-branchformer block configuration with subunits for the different e-branchformer parts
        """
        super().__init__()
        self.ff_1 = PositionwiseFeedForwardV1(cfg=cfg.ff_cfg)
        self.mhsa = MHSAV1(cfg=cfg.mhsa_cfg)
        self.cgmlp = ConvolutionalGatingMLPV1(model_cfg=cfg.cgmlp_cfg)
        self.merger = MergerV1(model_cfg=cfg.merger_cfg)
        self.ff_2 = PositionwiseFeedForwardV1(cfg=cfg.ff_cfg)
        self.final_layer_norm = torch.nn.LayerNorm(cfg.ff_cfg.input_dim)

    def forward(self, x: torch.Tensor, /, sequence_mask: torch.Tensor) -> torch.Tensor:
        """
        :param tensor: input tensor of shape [B, T, F]
        :param sequence_mask: mask tensor where 0 defines positions within the sequence and 1 outside, shape: [B, T]
        :return: torch.Tensor of shape [B, T, F]
        """
        x = 0.5 * self.ff_1(x) + x  #  [B, T, F]
        x_1 = self.mhsa(x, sequence_mask)  #  [B, T, F]
        x_2 = self.cgmlp(x)  #  [B, T, F]
        x = self.merger(x_1, x_2) + x  #  [B, T, F]
        x = 0.5 * self.ff_2(x) + x  #  [B, T, F]
        x = self.final_layer_norm(x)  # [B, T, F]
        return x


@dataclass
class EbranchformerEncoderV1Config(ModelConfiguration):
    """
    Attributes:
        num_layers: Number of e-branchformer layers in the e-branchformer encoder
        frontend: A pair of ConformerFrontend and corresponding config
        block_cfg: Configuration for EbranchformerBlockV1
    """

    num_layers: int

    # nested configurations
    frontend: ModuleFactoryV1
    block_cfg: EbranchformerBlockV1Config


class EbranchformerEncoderV1(nn.Module):
    """
    Implementation of the Branchformer with Enhanced merging (short e-branchformer), as in the original publication.
    The model consists of a frontend and a stack of N e-branchformer blocks.
    C.f. https://arxiv.org/pdf/2210.00077.pdf
    """

    def __init__(self, cfg: EbranchformerEncoderV1Config):
        """
        :param cfg: e-branchformer encoder configuration with subunits for frontend and e-branchformer blocks
        """
        super().__init__()

        self.frontend = cfg.frontend()
        self.module_list = torch.nn.ModuleList([EbranchformerBlockV1(cfg.block_cfg) for _ in range(cfg.num_layers)])

    def forward(self, data_tensor: torch.Tensor, sequence_mask: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        :param data_tensor: input tensor of shape [B, T', F']
        :param sequence_mask: mask tensor where 1 defines positions within the sequence and 0 outside, shape: [B, T']
        :return: (output, out_seq_mask)
            where output is torch.Tensor of shape [B, T, F],
            out_seq_mask is a torch.Tensor of shape [B, T]

        F': input feature dim, F: internal and output feature dim
        T': data time dim, T: down-sampled time dim (internal time dim)
        """
        x, sequence_mask = self.frontend(data_tensor, sequence_mask)  # [B, T, F]
        for module in self.module_list:
            x = module(x, sequence_mask)  # [B, T, F]

        return x, sequence_mask



__all__ = ["FeedForwardConfig", "FeedForwardModel"]

from dataclasses import dataclass
from functools import partial
from typing import Callable, Optional, Tuple, Union

import torch
from torch import nn
import torch.nn.functional as F

from i6_models.config import ModelConfiguration


@dataclass
class FeedForwardLayerV1Config(ModelConfiguration):
    """
    Attributes:
        input_dim: input feature dimension
        output_dim: output feature dimension
        dropout: dropout probability
        activation: activation function applied after linear computation
    """

    input_dim: int
    output_dim: int
    dropout: float
    activation: Union[nn.Module, Callable[[torch.Tensor], torch.Tensor]]

    def __post_init__(self):
        super().__post_init__()
        assert 0.0 <= self.dropout <= 1.0, "Dropout value must be a probability"


class FeedForwardLayerV1(nn.Module):
    """
    Simple feed-forward layer module consisting of:
        - linear
        - activation
        - dropout
    """

    def __init__(self, cfg: FeedForwardLayerV1Config):
        super().__init__()
        self.linear_ff = nn.Linear(in_features=cfg.input_dim, out_features=cfg.output_dim, bias=True)
        self.activation = cfg.activation
        self.dropout = nn.Dropout(cfg.dropout)

    def forward(
        self, tensor: torch.Tensor, sequence_mask: Optional[torch.Tensor] = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        :param tensor: shape [B,T,F], F=input_dim
        :param sequence_mask: shape [B,T]
        :return: shape [B,T,F'], F'=output_dim
        """
        tensor = self.linear_ff(tensor)  # [B,T,F]
        tensor = self.activation(tensor)  # [B,T,F]
        tensor = self.dropout(tensor)  # [B,T,F]
        return tensor, sequence_mask



"""
Usage:

Create ~/.pypirc with info:

    [distutils]
    index-servers =
        pypi

    [pypi]
    repository: https://upload.pypi.org/legacy/
    username: ...
    password: ...

(dep: pip3 install --user twine setuptools)
(cleanup: rm dist/* _setup_info_generated.py)

python3 setup.py sdist
twine upload dist/*.tar.gz

See also MANIFEST.in for included files.

For debugging this script:

python3 setup.py sdist
pip3 install --user dist/*.tar.gz -v
(Without -v, all stdout/stderr from here will not be shown.)

"""

import os
import shutil
from i6_models.__setup__ import get_version_str, debug_print_file


def main():
    """
    Setup main entry
    """
    # Do not use current time as fallback for the version anymore,
    # as this would result in a version which can be bigger than what we actually have,
    # so this would not be useful at all.
    long_version = get_version_str(verbose=True, fallback="0.0.1+setup-fallback-version", long=True)
    version = long_version[: long_version.index("+")]

    if os.environ.get("DEBUG", "") == "1":
        debug_print_file(".")
        debug_print_file("PKG-INFO")
        debug_print_file("pip-egg-info")
        debug_print_file("pip-egg-info/i6_models.egg-info")
        debug_print_file("pip-egg-info/i6_models.egg-info/SOURCES.txt")  # like MANIFEST

    if os.path.exists("PKG-INFO"):
        if os.path.exists("MANIFEST"):
            print("package_data, found PKG-INFO and MANIFEST")
            package_data = open("MANIFEST").read().splitlines() + ["PKG-INFO"]
        else:
            print("package_data, found PKG-INFO, no MANIFEST, use *")
            # Currently the setup will ignore all other data except in i6_models/.
            # At least make the version available.
            shutil.copy("PKG-INFO", "i6_models/")
            shutil.copy("_setup_info_generated.py", "i6_models/")
            # Just using package_data = ["*"] would only take files from current dir.
            package_data = []
            for root, dirs, files in os.walk("."):
                for file in files:
                    package_data.append(os.path.join(root, file))
    else:
        print("dummy package_data, does not matter, likely you are running sdist")
        with open("_setup_info_generated.py", "w") as f:
            f.write("version = %r\n" % version)
            f.write("long_version = %r\n" % long_version)
        package_data = ["MANIFEST", "_setup_info_generated.py"]

    from setuptools import setup

    authors = [
        dict(name="Christoph Lüscher", email="luescher@cs.rwth-aachen.de"),
        dict(name="Nick Rossenbach", email="rossenbach@cs.rwth-aachen.de"),
        dict(name="Benedikt Hilmes", email="hilmes@cs.rwth-aachen.de"),
        dict(name="Jingjing Xu", email="jxu@cs.rwth-aachen.de"),
        dict(name="Mohammad Zeineldeen", email="zeineldeen@cs.rwth-aachen.de"),
        dict(name="Albert Zeyer", email="zeyer@cs.rwth-aachen.de"),
        dict(name="Peter Vieting", email="vieting@cs.rwth-aachen.de"),
        dict(name="Simon Berger", email="sberger@cs.rwth-aachen.de"),
        dict(name="Eugen Beck", email="ebeck@apptek.com"),
        dict(name="Ping Zheng", email="p.zheng@apptek.com"),
        dict(name="Wilfried Michel", email="wmichel@apptek.com"),
    ]

    setup(
        name="i6_models",
        version=version,
        packages=["i6_models"],
        include_package_data=True,
        package_data={"i6_models": package_data},  # filtered via MANIFEST.in
        description="A collection of PyTorch NN models and parts.",
        author=", ".join(d["name"] for d in authors),
        author_email=", ".join(d["email"] for d in authors),
        url="https://github.com/rwth-i6/i6_models/",
        license="Mozilla Public License 2.0",
        long_description=open("README.md").read(),
        long_description_content_type="text/markdown",
        # https://pypi.org/classifiers/
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
            "Operating System :: OS Independent",
        ],
    )


if __name__ == "__main__":
    main()



"""
Provides the base class for configurations, model configurations can be derived from this base class e.g:

@dataclass
class ExampleConfig(ModelConfiguration):
    hidden_dim: int = 256
    name: str = "Example Configuration"

class ExampleModule(Module):
    __init__(self, cfg: ExampleConfig)
        self.hidden_dim = cfg.hidden_dim

This config can then be used in the construction of the model to provide parameters.
Similar approach as done in Fairseq: https://github.com/facebookresearch/fairseq/blob/main/fairseq/dataclass/configs.py
"""

from __future__ import annotations
from dataclasses import dataclass, fields
import typeguard
from torch import nn
from typing import Generic, TypeVar, Type
import inspect


@dataclass
class ModelConfiguration:
    """
    Base dataclass for configuration of different primitives, parts and assemblies.
    Enforces type checking for the creation of derived configs.
    """

    def _validate_types(self) -> None:
        for field in fields(type(self)):
            try:
                if isinstance(field.type, str):
                    # Later Python versions use forward references,
                    # and then field.type is a str object.
                    # This needs to be resolved to the type itself.
                    # At the moment is does not look like there is typeguard support for it, thus manual fixes.
                    # C.f. https://github.com/agronholm/typeguard/issues/358
                    from typeguard._functions import check_type_internal
                    from typeguard._memo import TypeCheckMemo
                    from typing import ForwardRef
                    import sys

                    cls_globals = vars(sys.modules[self.__class__.__module__])
                    memo = TypeCheckMemo(globals=cls_globals, locals=cls_globals)
                    t = ForwardRef(field.type)
                    check_type_internal(getattr(self, field.name), t, memo)
                else:
                    typeguard.check_type(getattr(self, field.name), field.type)
            except typeguard.TypeCheckError as exc:
                raise typeguard.TypeCheckError(f'In field "{field.name}" of "{type(self)}": {exc}')

    def __post_init__(self) -> None:
        self._validate_types()


ConfigType = TypeVar("ConfigType", bound=ModelConfiguration)
ModuleType = TypeVar("ModuleType", bound=nn.Module)


@dataclass
class ModuleFactoryV1(Generic[ConfigType, ModuleType]):
    """
    Dataclass for a combination of a Subassembly/Part and the corresponding configuration.
    Also provides a function to construct the corresponding object through this dataclass
    """

    module_class: Type[ModuleType]
    cfg: ConfigType

    def __call__(self) -> ModuleType:
        """Constructs an instance of the given module class"""
        return self.module_class(self.cfg)

    def __post_init__(self) -> None:
        # Check typing of module_class and cfg, i.e. make sure that "self.module_class(self.cfg)" is a valid call.
        parameters = inspect.signature(self.module_class).parameters.values()
        assert len(parameters) >= 1
        parameter_iter = iter(parameters)

        # 1. Check that the first parameter is either not annotated or the annotation matches the type of self.cfg
        cfg_parameter = next(parameter_iter)
        if cfg_parameter.annotation is not inspect.Parameter.empty:
            typeguard.check_type(self.cfg, cfg_parameter.annotation)

        # 2. Check that all other parameters have default values
        for parameter in parameter_iter:
            assert parameter.default is not inspect.Parameter.empty
